@@ -1,5 +1,5 @@
 /* ====================================================================
- * Copyright (c) 1995, 1996, 1997 Ben Laurie.  All rights reserved.
+ * Copyright (c) 1995, 1996, 1997, 1998 Ben Laurie.  All rights reserved.
  *
  * Redistribution and use in source and binary forms, with or without
  * modification, are permitted provided that the following conditions
@@ -43,12 +43,14 @@
  *
  * This software consists of patches to the Apache HTTP server interfacing it
  * to SSLeay.
- * For more information on Apache-SSL, contact Ben Laurie <ben@algroup.co.uk>.
+ * For more information on Apache-SSL see http://www.apache-ssl.org.
  *
  * For more information on Apache see http://www.apache.org.
  *
  * For more information on SSLeay see http://www.psy.uq.oz.au/~ftp/Crypto/.
  *
+ * Ben Laurie <ben@algroup.co.uk>.
+ *
  */
 
 #include <stdio.h>
@@ -91,16 +93,15 @@ use /u/ben/work/scuzzy-ssleay8/ssl /u/ben/work/scuzzy-ssleay8/crypto/bio
 # define CACHE_SESSIONS		TRUE
 #endif
 
-/* define this true when SSLeay can do an SSL_free without falling over */
-#define FREE_SESSION	FALSE
+#define FREE_SESSION	TRUE	/* Don't set this FALSE! */
 
 /* define this to debug the verification stuff */
 #define DEBUG_VERIFY	FALSE
 
 #if CACHE_SESSIONS
 # include "gcache.h"
-static const char *szCacheServerPort;
-static char *szCacheServerPath;
+static const char *s_szCacheServerPort;
+static char *s_szCacheServerPath;
 
 #define DEBUG_SESSIONS		0
 
@@ -145,9 +146,7 @@ typedef struct
     char *szKeyFile;
     char *szCACertificatePath;
     char *szCACertificateFile;
-    char *szLogFile;
     char *szReqCiphers;
-    FILE *fileLogFile;
     int nVerifyDepth;
     VerifyType nVerifyClient;
 
@@ -172,9 +171,9 @@ typedef struct
     BOOL bSSLRequired;
     } SSLDirConfigRec;
 
-extern module apache_ssl_module;
+MODULE_VAR_EXPORT module apache_ssl_module;
 
-static conn_rec *g_pCurrentConnection;
+static conn_rec *s_pCurrentConnection;
 
 static const char *SSLRequireCipher(cmd_parms *cmd,SSLDirConfigRec *rec,
 				    char *cipher)
@@ -229,14 +228,14 @@ static int SSLCheckCipher(request_rec *r)
     if(ap_table_get(rec->tbRequiredCiphers,cipher))
 	return OK;
 
-    if(is_empty_table(rec->tbRequiredCiphers))
+    if(ap_is_empty_table(rec->tbRequiredCiphers))
 	return OK;
     else
 	{
 	char *buf;
 	
-	buf=ap_pstrcat(r->pool,"Cipher ",cipher," is not on the permitted list",
-		    NULL);
+	buf=ap_pstrcat(r->pool,"Cipher ",cipher,
+		       " is not on the permitted list",NULL);
 	ap_log_reason(buf,r->filename,r);
 	return FORBIDDEN;
 	}
@@ -248,7 +247,7 @@ static void StrUpper(char *s)
 	*s=toupper(*s);
     }
 
-void ExpandCert(table *pEnv,char *szPrefix,char *szCert)
+static void ExpandCert(table *pEnv,char *szPrefix,char *szCert)
     {
     /* Use ap_pstrintf when we get to 1.3 */
     char var[HUGE_STRING_LEN];
@@ -323,7 +322,7 @@ static struct
     { NULL, 0, 0 }
     };
 
-int SSLFixups(request_rec *r)
+static int SSLFixups(request_rec *r)
     {
     table *e=r->subprocess_env;
     int keysize=0;
@@ -379,17 +378,17 @@ int SSLFixups(request_rec *r)
     return DECLINED;
     }
 
-static void SSLLogError(server_rec *s)
+static void SSLLogError(server_rec *s,int nLevel,const char *szErr)
     {
     unsigned long l;
     char buf[MAX_STRING_LEN];
 
-    /* Print out error messages */
-
+    if(szErr)
+	ap_log_error(APLOG_MARK,nLevel|APLOG_NOERRNO,s,"%s",szErr);
     while((l=ERR_get_error()))
 	{
 	ERR_error_string(l,buf);
-	ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,s,"%s",buf);
+	ap_log_error(APLOG_MARK,nLevel|APLOG_NOERRNO,s,"%s",buf);
 	}
     }
 
@@ -418,9 +417,7 @@ static void *CreateSSLServerConfig(pool *p,server_rec *s)
     rec->szKeyFile=NULL;
     rec->szCACertificatePath=NULL;
     rec->szCACertificateFile=NULL;
-    rec->szLogFile=NULL;
     rec->szReqCiphers=NULL;
-    rec->fileLogFile=NULL;
     rec->nVerifyDepth=0;
     rec->nVerifyClient=VERIFY_UNSET;	/* Will become VERIFY_NONE */
     rec->px509Certificate=NULL;
@@ -448,7 +445,6 @@ static void *MergeSSLServerConfig(pool *p,void *basev,void *addv)
     MergeString(szKeyFile);
     MergeString(szCACertificatePath);
     MergeString(szCACertificateFile);
-    MergeString(szLogFile);
     MergeInt(nVerifyDepth);
     Merge(nVerifyClient,VERIFY_UNSET);
     MergeBool(bFakeBasicAuth);
@@ -483,19 +479,21 @@ static void *MergeSSLDirConfig(pool *p,void *basev,void *addv)
     return new;
     }
 
-static void InitSSL()
+static void InitSSL(server_rec *s)
     {
     char *CAfile=NULL,*CApath=NULL;
 
 #if CACHE_SESSIONS
-    if(!szCacheServerPort)
+    if(!s_szCacheServerPort)
 	{
-	fprintf(stderr,"Required SSLCacheServerPort missing\n");
+	ap_log_error(APLOG_MARK,APLOG_CRIT|APLOG_NOERRNO,s,
+		"Required SSLCacheServerPort missing");
 	exit(1);
 	}
-    if(!szCacheServerPath)
+    if(!s_szCacheServerPath)
 	{
-	fprintf(stderr,"Required SSLCacheServerPath missing\n");
+	ap_log_error(APLOG_MARK,APLOG_CRIT|APLOG_NOERRNO,s,
+		     "Required SSLCacheServerPath missing");
 	exit(1);
 	}
 #endif
@@ -528,16 +526,18 @@ static RSA *TmpRSACallback(SSL *pSSL,int nExport)
 #endif
 
 #if SSLEAY_VERSION_NUMBER >= 0x0800
-int ApacheSSLVerifyCallback(int ok,X509_STORE_CTX *ctx)
+static int ApacheSSLVerifyCallback(int ok,X509_STORE_CTX *ctx)
     {
     X509 *xs=X509_STORE_CTX_get_current_cert(ctx);
     int depth=X509_STORE_CTX_get_error_depth(ctx);
     int error=X509_STORE_CTX_get_error(ctx);
 #else
-int ApacheSSLVerifyCallback(int ok,X509 *xs,X509 *xi,int depth,int error)
+static int ApacheSSLVerifyCallback(int ok,X509 *xs,X509 *xi,int depth,
+				   int error)
     {
 #endif
-    SSLConfigRec *pConfig=ap_get_module_config(g_pCurrentConnection->server->module_config,&apache_ssl_module);
+    SSLConfigRec *pConfig=ap_get_module_config(s_pCurrentConnection->server->module_config,&apache_ssl_module);
+    server_rec *pServer=s_pCurrentConnection->server;
 
 #if DEBUG_VERIFY
       {
@@ -551,11 +551,12 @@ int ApacheSSLVerifyCallback(int ok,X509 *xs,X509 *xi,int depth,int error)
 
       if(s == NULL)
 	  {
-	  ERR_print_errors_fp(pConfig->fileLogFile);
+	  SSLLogError(pServer,APLOG_ERR,"can't get subject name");
 	  return(0);
 	  }
 
-      fprintf(pConfig->fileLogFile,"depth=%d %s\n",depth,s);
+      ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,"depth=%d %s",
+		   depth,s);
       free(s);
       }
 #endif
@@ -566,7 +567,8 @@ int ApacheSSLVerifyCallback(int ok,X509 *xs,X509 *xi,int depth,int error)
 
 	if(pConfig->nVerifyClient == VERIFY_OPTIONAL_NO_CA)
 	    {
-	    fprintf(pConfig->fileLogFile,"no issuer, returning OK\n");
+	    ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,
+			 "no issuer, returning OK");
 	    return TRUE;
 	    }
 #if SSLEAY_VERSION_NUMBER < 0x0800
@@ -576,29 +578,29 @@ int ApacheSSLVerifyCallback(int ok,X509 *xs,X509 *xi,int depth,int error)
 #endif
 	if(s == NULL)
 	    {
-	    fprintf(pConfig->fileLogFile,"verify error\n");
-	    ERR_print_errors_fp(pConfig->fileLogFile);
-	    SSLLogError(g_pCurrentConnection->server);
+	    SSLLogError(pServer,APLOG_ERR,"verify error");
 	    return(0);
 	    }
-	fprintf(pConfig->fileLogFile,"issuer= %s\n",s);
+	ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,"issuer=%s",
+		     s);
 	free(s);
 	}
     if(!ok)
 	{
-	fprintf(pConfig->fileLogFile,"verify error:num=%d:%s\n",error,
-		X509_verify_cert_error_string(error));
-	SSLLogError(g_pCurrentConnection->server);
-	g_pCurrentConnection->client->szClientX509=NULL;
+	ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,pServer,
+		      "verify error:num=%d:%s",error,
+		      X509_verify_cert_error_string(error));
+	SSLLogError(s_pCurrentConnection->server,APLOG_ERR,NULL);
+	s_pCurrentConnection->client->szClientX509=NULL;
 	}
     if(depth >= pConfig->nVerifyDepth)
 	{
-	fprintf(pConfig->fileLogFile,"Verify depth exceeded\n");
 	ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,
-		     g_pCurrentConnection->server,"Verify depth exceeded");
+		     pServer,"Verify depth exceeded");
 	ok=0;
 	}
-    fprintf(pConfig->fileLogFile,"verify return:%d\n",ok);
+    ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,pServer,
+		 "verify return:%d",ok);
     return(ok);
     }
 
@@ -636,7 +638,7 @@ static void SendSessionToServer(SSL_SESSION *pSession)
     if(LocalCacheFind(pSession->session_id,pSession->session_id_length))
 	return;
 
-    pConfig=ap_get_module_config(g_pCurrentConnection->server->module_config,
+    pConfig=ap_get_module_config(s_pCurrentConnection->server->module_config,
 				 &apache_ssl_module);
     tExpiresAt=time(NULL)+pConfig->tSessionCacheTimeout;
 
@@ -735,13 +737,14 @@ static void GetCertificateAndKey(server_rec *s,SSLConfigRec *pConfig)
 
     if(pConfig->px509Certificate)
 	{
-	fprintf(stderr,"Attempt to reinitialise SSL for server %s\n",
-		s->server_hostname);
+	ap_log_error(APLOG_MARK,APLOG_NOERRNO|APLOG_CRIT,s,
+		     "Attempt to reinitialise SSL for server %s",
+		     s->server_hostname);
 	exit(1);
 	}
 
-    printf("Reading certificate and key for server %s:%d\n",
-	   s->server_hostname,s->port);
+    fprintf(stderr,"Reading certificate and key for server %s:%d\n",
+	    s->server_hostname,s->port);
 
 #if SSLEAY_VERSION_NUMBER < 0x0800
     pConfig->pSSLCtx=SSL_CTX_new();
@@ -759,9 +762,8 @@ static void GetCertificateAndKey(server_rec *s,SSLConfigRec *pConfig)
 	{
 	if(!SSL_CTX_set_cipher_list(pConfig->pSSLCtx,pConfig->szReqCiphers))
 	    {
-	    fprintf(stderr,"unable to set ciphers\n");
-	    ERR_print_errors_fp(stderr);
-	    SSLLogError(s);
+	    SSLLogError(s,APLOG_CRIT,"unable to set ciphers");
+	    exit(1);
 	    }
 	}
 
@@ -772,8 +774,7 @@ static void GetCertificateAndKey(server_rec *s,SSLConfigRec *pConfig)
 	)
        || !SSL_CTX_set_default_verify_paths(pConfig->pSSLCtx))
 	{
-	fprintf(stderr,"error setting verify locations\n");
-	ERR_print_errors_fp(stderr);
+	SSLLogError(s,APLOG_CRIT,"error setting verify locations");
 	exit(1);
 	}
 
@@ -785,8 +786,9 @@ static void GetCertificateAndKey(server_rec *s,SSLConfigRec *pConfig)
 	f=fopen(szPath,"r");
 	if(!f)
 	    {
-	    fprintf(stderr,"Can't open certificate file %s, nor %s\n",
-		    pConfig->szCertificateFile,szPath);
+	    ap_log_error(APLOG_MARK,APLOG_CRIT|APLOG_NOERRNO,s,
+			 "Can't open certificate file %s, nor %s",
+			 pConfig->szCertificateFile,szPath);
 	    exit(1);
 	    }
 	}
@@ -797,8 +799,9 @@ static void GetCertificateAndKey(server_rec *s,SSLConfigRec *pConfig)
 
     if(!PEM_read_X509(f,&pConfig->px509Certificate,NULL))
 	{
-	fprintf(stderr,"Error reading server certificate file %s: ",szPath);
-	ERR_print_errors_fp(stderr);
+	ap_log_error(APLOG_MARK,APLOG_CRIT|APLOG_NOERRNO,s,
+		     "Error reading server certificate file %s",szPath);
+	SSLLogError(s,APLOG_CRIT,NULL);
 	exit(1);
 	}
     fclose(f);
@@ -814,16 +817,16 @@ static void GetCertificateAndKey(server_rec *s,SSLConfigRec *pConfig)
     f=fopen(szPath,"r");
     if(!f)
 	{
-	fprintf(stderr,"Can't open key file ");
-	perror(szPath);
+	ap_log_error(APLOG_MARK,APLOG_CRIT,s,"Can't open key file %s",szPath);
 	exit(1);
 	}
 
     pConfig->prsaKey=RSA_new();
     if(!PEM_read_RSAPrivateKey(f,&pConfig->prsaKey,NULL))
 	{
-	fprintf(stderr,"Error reading private key file %s: ",szPath);
-	ERR_print_errors_fp(stderr);
+	ap_log_error(APLOG_MARK,APLOG_CRIT,s,
+		     "Error reading private key file %s: ",szPath);
+	SSLLogError(s,APLOG_CRIT,NULL);
 	exit(1);
 	}
     fclose(f);
@@ -851,7 +854,9 @@ static int CacheLauncher(void *cmd,child_info *pInfo)
     ap_error_log2stderr(s_pServer);
     dup2(STDERR_FILENO,STDOUT_FILENO);
     sprintf(a1,"%d",ap_user_id);
-    execl(cmd,"gcache",a1,szCacheServerPort,NULL);
+    ap_cleanup_for_exec();
+    execl(cmd,"gcache",a1,s_szCacheServerPort,NULL);
+    perror(cmd);
     assert(0);
     return 1;
     }
@@ -861,8 +866,8 @@ static void LaunchCache(server_rec *s,pool *pPool)
     int pid;
 
     s_pServer=s;
-    printf("Launching... %s\n",szCacheServerPath);
-    pid=ap_spawn_child(pPool,CacheLauncher,szCacheServerPath,kill_after_timeout,NULL,NULL,NULL);
+    printf("Launching... %s\n",s_szCacheServerPath);
+    pid=ap_spawn_child(pPool,CacheLauncher,s_szCacheServerPath,kill_always,NULL,NULL,NULL);
     printf("pid=%d\n",pid);
     }
 #endif
@@ -873,10 +878,27 @@ static void InitSSLServer(server_rec *s,pool *p)
     static BOOL bFirstTime=TRUE;
     static int nStashStdin,nStashStderr,nSaveStdin,nSaveStderr;
     int n;
+    server_rec *pMainServer=s;
+
+    /* If you were going to make Apache-SSL add a version string by the new
+       method, here's where to do it. However, it doesn't seem right to me to
+       do that, since if APACHE_SSL is defined, the server is modified,
+       regardless of whether this module is included. - Ben 3/8/98
+
+       ap_add_version_component(APACHE_SSL_VERSION);
+
+       */
+
+#ifdef SHARED_MODULE
+    /* Allegedly this is only run once when included as a DSO (though the value
+     of doing that is questionable, since the rest of the code doesn't work
+     without this module).
+     */
+    bFirstTime=FALSE;
+#endif
 
     if(bFirstTime)
 	{
-	printf("Skip first time initialisation\n");
 	bFirstTime=FALSE;
 	nStashStdin=dup(0);
 	assert(nStashStdin > 0);
@@ -898,10 +920,7 @@ static void InitSSLServer(server_rec *s,pool *p)
     n=dup(nStashStderr);
     assert(n == 2);
 
-    InitSSL();
-#if CACHE_SESSIONS
-    LaunchCache(s,p);
-#endif
+    InitSSL(s);
     for ( ; s ; s=s->next)
 	{
 	SSLConfigRec *pConfig=ap_get_module_config(s->module_config,
@@ -922,14 +941,16 @@ static void InitSSLServer(server_rec *s,pool *p)
 	    {
 	    if(!s->port)
 		s->port=DEFAULT_HTTP_PORT;
-	    fprintf(stderr,"SSL disabled for server %s:%d\n",
-		    s->server_hostname,s->port);
+	    ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,s,
+			 "SSL disabled for server %s:%d\n",
+			 s->server_hostname,s->port);
 	    continue;
 	    }
 #if CACHE_SESSIONS
 	if(!pConfig->tSessionCacheTimeout)
 	    {
-	    fprintf(stderr,"SSLSessionCacheTimeout not set\n");
+	    ap_log_error(APLOG_MARK,APLOG_CRIT|APLOG_NOERRNO,s,
+			 "SSLSessionCacheTimeout not set");
 	    exit(1);
 	    }
 #endif
@@ -938,32 +959,23 @@ static void InitSSLServer(server_rec *s,pool *p)
 
 	if(!pConfig->szCertificateFile)
 	    {
-	    fprintf(stderr,"No SSL Certificate set for server %s:%d\n",
-		    s->server_hostname,s->port);
+	    ap_log_error(APLOG_MARK,APLOG_CRIT|APLOG_NOERRNO,s,
+			 "No SSL Certificate set for server %s:%d",
+			 s->server_hostname,s->port);
 	    exit(1);
 	    }
 	if(pConfig->nVerifyClient < 0 || pConfig->nVerifyClient > VERIFY_OPTIONAL_NO_CA)
 	    {
-	    fprintf(stderr,"Bad value for SSLVerifyClient (%d)\n",pConfig->nVerifyClient);
+	    ap_log_error(APLOG_MARK,APLOG_CRIT|APLOG_NOERRNO,s,
+			 "Bad value for SSLVerifyClient (%d)",
+			 pConfig->nVerifyClient);
 	    exit(1);
 	    }
-	if(!pConfig->szLogFile)
-	    {
-	    fprintf(stderr,"Required SSLLogFile missing\n");
-	    exit(1);
-	    }
-
-	szLogFile=ap_server_root_relative(p,pConfig->szLogFile);
-	pConfig->fileLogFile=fopen(szLogFile,"a");
-	if(!pConfig->fileLogFile)
-	    {
-	    perror(pConfig->szLogFile);
-	    exit(1);
-	    }
-	setbuf(pConfig->fileLogFile,NULL);
-    
 	GetCertificateAndKey(s,pConfig);
 	}
+#if CACHE_SESSIONS
+    LaunchCache(pMainServer,p);
+#endif
     close(0);
     n=dup(nSaveStdin);
     assert(n == 0);
@@ -976,26 +988,33 @@ static void InitSSLServer(server_rec *s,pool *p)
     }
 
 #if CACHE_SESSIONS
-static const char *SSLCacheServerPort(cmd_parms *cmd,char *struct_ptr,char *arg)
+static const char *SSLCacheServerPort(cmd_parms *cmd,char *struct_ptr,
+				      char *arg)
     {
     if(strspn(arg,"0123456789") == strlen(arg))
-	szCacheServerPort=arg;
+	s_szCacheServerPort=arg;
     else
-	szCacheServerPort=ap_server_root_relative(cmd->pool,arg);
-    InitGlobalCache(szCacheServerPort);
+	s_szCacheServerPort=ap_server_root_relative(cmd->pool,arg);
+    InitGlobalCache(s_szCacheServerPort);
 
     return NULL;
     }
 
-static const char *SSLCacheServerPath(cmd_parms *cmd,char *struct_ptr,char *arg)
+static const char *SSLCacheServerPath(cmd_parms *cmd,char *struct_ptr,
+				      char *arg)
     {
-    szCacheServerPath=ap_server_root_relative(cmd->pool,arg);
+    s_szCacheServerPath=ap_server_root_relative(cmd->pool,arg);
 
     return NULL;
     }
 #endif
     
 
+static const char *SSLLogFile(cmd_parms *cmd,char *struct_ptr,char *arg)
+    {
+    return "The SSLLogFile directive is no longer used. All logging information now goes to the error log.";
+    }
+
 static const char *set_server_string_slot (cmd_parms *cmd,char *struct_ptr,char *arg)
     {
     /* This one should be pretty generic... */
@@ -1069,9 +1088,6 @@ command_rec ssl_cmds[]=
     { "SSLFakeBasicAuth",set_server_bool_slot,
       (void *)XtOffsetOf(SSLConfigRec,bFakeBasicAuth),RSRC_CONF,NO_ARGS,
       "Translate client X509 into a user name" },
-    { "SSLLogFile",set_server_string_slot,
-      (void *)XtOffsetOf(SSLConfigRec,szLogFile),RSRC_CONF,TAKE1,
-      "Place to dump all SSL messages that have no better home" },
     { "SSLRequiredCiphers",set_server_string_slot,
       (void *)XtOffsetOf(SSLConfigRec,szReqCiphers),RSRC_CONF,TAKE1,
       "Colon-delimited list of required ciphers" },
@@ -1091,6 +1107,8 @@ command_rec ssl_cmds[]=
     { "SSLCacheServerPath",SSLCacheServerPath,NULL,RSRC_CONF,TAKE1,
       "Path of the global cache server" },
 #endif
+    { "SSLLogFile",SSLLogFile,NULL,RSRC_CONF,TAKE1,
+      "Defunct SSL log file - DO NOT USE" },
     { NULL },
     };
 
@@ -1172,23 +1190,19 @@ static BOOL ApacheSSLSetCertStuff(conn_rec *conn)
 	{
 	if(SSL_use_certificate(con,pConfig->px509Certificate) <= 0)
 	    {
-	    fprintf(pConfig->fileLogFile,"unable to set certificate\n");
-	    ERR_print_errors_fp(pConfig->fileLogFile);
-	    SSLLogError(conn->server);
+	    SSLLogError(conn->server,APLOG_CRIT,"unable to set certificate");
 	    return FALSE;
 	    }
 	if(SSL_use_RSAPrivateKey(con,pConfig->prsaKey) <= 0)
 	    {
-	    fprintf(pConfig->fileLogFile,"unable to set private key\n");
-	    ERR_print_errors_fp(pConfig->fileLogFile);
-	    SSLLogError(conn->server);
+	    SSLLogError(conn->server,APLOG_CRIT,"unable to set private key");
 	    return FALSE;
 	    }
 	}
     return TRUE;
     }
 
-void ApacheSSLSetupVerify(conn_rec *conn)
+static void ApacheSSLSetupVerify(conn_rec *conn)
     {
     SSLConfigRec *pConfig=ap_get_module_config(conn->server->module_config,
 					       &apache_ssl_module);
@@ -1196,7 +1210,7 @@ void ApacheSSLSetupVerify(conn_rec *conn)
 
     conn->client->szClientX509=NULL;
 
-    g_pCurrentConnection=conn;
+    s_pCurrentConnection=conn;
 
     conn->client->nVerifyError=X509_V_OK;
 
@@ -1206,7 +1220,6 @@ void ApacheSSLSetupVerify(conn_rec *conn)
 #endif
     if(!ApacheSSLSetCertStuff(conn))
 	{
-	fprintf(pConfig->fileLogFile,"ApacheSSLSetCertStuff failed\n");
 	ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,conn->server,
 		     "ApacheSSLSetCertStuff failed");
 	exit(1);
@@ -1217,20 +1230,14 @@ void ApacheSSLSetupVerify(conn_rec *conn)
     }
 
 #if FREE_SESSION
-/*
-To make this work, SSL_free needs to be patched. Remove from SSL_free:
-
-	if (s->ctx) SSL_CTX_free(s->ctx);
-
-but there's more, which I haven't found.
-*/
-
-static void CloseSSL(void *_ssl)
+static void CloseSSL(void *_conn)
     {
-    SSL *ssl=_ssl;
-printf("Closing\n");
-    SSL_free(_ssl);
-printf("Closed\n");
+    conn_rec *conn=_conn;
+    SSL *ssl=conn->client->ssl;
+
+    SSL_free(ssl);
+    /*    ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,conn->server,
+		 "Closed");*/
     }
 #endif
 
@@ -1249,12 +1256,12 @@ int ApacheSSLSetupConnection(conn_rec * conn)
 	return TRUE;
 	}
     
-    g_pCurrentConnection=conn;
+    s_pCurrentConnection=conn;
 
     fb->ssl=SSL_new(pConfig->pSSLCtx);
     SSL_set_fd(fb->ssl,fb->fd);
 #if FREE_SESSION
-    register_cleanup(conn->pool,fb->ssl,CloseSSL,CloseSSL);
+    ap_register_cleanup(conn->pool,conn,CloseSSL,CloseSSL);
 #endif
 
     ApacheSSLSetupVerify(conn);
@@ -1264,18 +1271,14 @@ int ApacheSSLSetupConnection(conn_rec * conn)
 	int ret=SSL_accept(fb->ssl);
 	if (ret <= 0)
 	    {
-	    ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,srvr,
-			 "SSL_Accept failed");
-	    SSLLogError(srvr);
+	    SSLLogError(srvr,APLOG_ERR,"SSL_accept failed");
 	    fb->flags |= B_EOF | B_EOUT;
 	    return FALSE;
 	    }
 
 	if(conn->client->nVerifyError != X509_V_OK)
 	    {
-	    ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,conn->server,
-			 "Verification failed");
-	    SSLLogError(srvr);
+	    SSLLogError(srvr,APLOG_ERR,"Verification failed");
 	    fb->flags |= B_EOF | B_EOUT;
 	    return FALSE;
 	    }
@@ -1295,24 +1298,22 @@ int ApacheSSLSetupConnection(conn_rec * conn)
 
 	if(pConfig->nVerifyClient == VERIFY_REQUIRE && !conn->client->szClientX509)
 	    {
-	    ap_log_error(APLOG_MARK,APLOG_ERR|APLOG_NOERRNO,conn->server,
-			 "No client certificate");
-	    SSLLogError(conn->server);
+	    SSLLogError(conn->server,APLOG_ERR,"No client certificate");
 	    return 0;
 	    }
-	fprintf(pConfig->fileLogFile,"CIPHER is %s\n",SSL_get_cipher(conn->client->ssl));
-	fflush(pConfig->fileLogFile);
+	ap_log_error(APLOG_MARK,APLOG_DEBUG|APLOG_NOERRNO,srvr,"CIPHER is %s",
+		     SSL_get_cipher(conn->client->ssl));
 	}
 
     /* This should be safe.... so I'll use it */
     SSL_set_read_ahead(fb->ssl,1);
 
-    g_pCurrentConnection=NULL;
+    s_pCurrentConnection=NULL;
 
     return TRUE;
     }
 
-module apache_ssl_module =
+module MODULE_VAR_EXPORT apache_ssl_module =
     {
     STANDARD_MODULE_STUFF,
     InitSSLServer,		/* initializer */
